#
# dumpDCWorkflow.py
# (c) inQual 2001
# written by Sebastien Bigaret
#
"""
Through this script you can either:

  - dump a (e.g. TTW defined) DCWorkflow into a python module so that
    the latter can be integrated in a Zope Product, making it
    available as a workflow type in CMF --just like default
    DCWorkflows: Web-configurable workflow [Classic] and
    Web-configurable workflow [Revision 2].

  - exports the 'scripts' folder within a given DCWorkflow

  - import a previously 'scripts' folder into a given DCWorkflow. Please note
    that this operation is destructive: all previously defined scripts within
    the DCWorkflow will be deleted.

See http://cmf.zope.org/Members/Big/dumpDCWorkflowHowto/view for details
"""
__version__='$Revision: 25723 $'[11:-2]

from pprint import pformat
from textwrap import wrap as _wrap
from Acquisition import aq_base
from StringIO import StringIO
from DateTime import DateTime
from Globals import MessageDialog
from Products.DCWorkflow.DCWorkflow import DCWorkflowDefinition

triple_quotes='"""'

def wrap(v, indent=0):
    """Wrap all lines but the first up to `indent` columns
    """
    width = 76 - indent
    wrapped = False
    if isinstance(v, (str, unicode)):
        w = tuple(_wrap(v, width=width))
        if len(w) > 1:
            wrapped = True
            v = []
            for l in w[:-1]:
                v.append(l + ' ')
            v.append(w[-1])
            v = tuple(v)
        elif len(w) == 1:
            v = w[0]
    s = pformat(v)
    r = []
    lines = filter(None, [l.strip() for l in s.splitlines()])
    if not lines:
        return s
    first = lines[0]
    if wrapped:
        first = first.strip(',')
    r.append(first)
    for l in lines[1:]:
        if wrapped:
            l = l.strip(',')
        r.append(" " * indent + l)
    return '\n'.join(r)

def quote(s):
    """Find best quote to apply
    """
    return pformat(s)

def DCWorkflow_importExport(self, aDCWorkflowId='survey_workflow',
                            importFilename='survey_workflow.py',
                            set_owner=1, REQUEST=None):
    """Export/import facilities for DCWorkflow.

    If REQUEST has key 'exportScripts', it exports the 'scripts' Folder.
    If REQUEST has key 'importScripts', it imports the 'scripts' Folder.
    Otherwise, it dumps the workflow 'aDCWorkflowId' into a python module.

    Returns the python file corresponding to the workflow identified by
    the supplied id.
    """
    wf = self.portal_workflow.getWorkflowById(aDCWorkflowId)

    if wf is None:
        msg='Workflow id:'+aDCWorkflowId+' not found!'
        return MessageDialog(title=msg, message=msg, action='./manage_main',)

    if type(aq_base(wf))!=type(DCWorkflowDefinition('test')):
        msg='This script only handles DCWorkflow workflows',
        return MessageDialog(title=msg, message=msg, action='./manage_main',)

    if REQUEST and hasattr(REQUEST, 'exportScripts'):
        export=exportWorkflowScripts(wf, REQUEST)
        return export

    if REQUEST and hasattr(REQUEST, 'importScripts'):
        importWorkflowScripts(self, wf, REQUEST)
        return MessageDialog(title='Scripts successfully imported!',
                             message='Scripts successfully imported!',
                             action ='./portal_workflow/%s/scripts/manage_main'%wf.id,)

    _result= StringIO()
    _result.write('#\n')

    dumpDCwf(wf, _result)

    _result.flush()
    _ret = _result.getvalue()
    _result.close()
    return _ret

def importWorkflowScripts(self, wf, REQUEST):
    """Import the 'scripts' folder within the supplied workflow.
    """
    importFile = REQUEST['importFilename']
    # First delete the 'scripts' ContainerTab
    wf._delObject('scripts')
    wf.manage_importObject(importFile, REQUEST, REQUEST['set_owner'])

def exportWorkflowScripts(wf, REQUEST):
    """Export the scripts within the supplied wf.
    """
    export=wf.scripts.manage_exportObject(id='', download='Yes!',
                                          RESPONSE=REQUEST.RESPONSE)
    REQUEST.RESPONSE.setHeader('Content-Disposition',
                               'inline;filename=%s_scripts.zexp' % (wf.id,))
    return export

def dumpDCwf(wf, aStringIO):
    """Dump the supplied workflow into the stringIO.
    """
    writeHeader(wf, aStringIO)
    dumpMainVariables(wf, aStringIO)
    dumpStates(wf, aStringIO)
    dumpTransitions(wf, aStringIO)
    dumpVariables(wf, aStringIO)
    dumpWorklists(wf, aStringIO)
    dumpScripts(wf, aStringIO)
    writeFooter(wf, aStringIO)

def getSortedIds(container):
    l = list(container.objectIds())
    l.sort()
    return l

def getSortedObjects(container):
    l = list(container.objectValues())
    def id_cmp(o1, o2):
        return cmp(o1.getId(), o2.getId())

    l.sort(id_cmp)
    return l

def dumpMainVariables(wf, aStrIO):
    indent = 14
    aStrIO.write("""
from Products.CMFCore.WorkflowTool import addWorkflowFactory
from Products.DCWorkflow.DCWorkflow import DCWorkflowDefinition
from Products.PythonScripts.PythonScript import PythonScript
from Products.ExternalMethod.ExternalMethod import ExternalMethod

def setup_%s(wf):
    \"\"\"Setup the workflow
    \"\"\"
    wf.setProperties(title=%s)

    for s in %s:
        wf.states.addState(s)
    for t in %s:
        wf.transitions.addTransition(t)
    for v in %s:
        wf.variables.addVariable(v)
    for l in %s:
        wf.worklists.addWorklist(l)
    for p in %s:
        wf.addManagedPermission(p)

""" % (wf.id,
       wrap(wf.title, indent=28),
       wrap(tuple(getSortedIds(wf.states)), indent),
       wrap(tuple(getSortedIds(wf.transitions)), indent),
       wrap(tuple(getSortedIds(wf.variables)), indent),
       wrap(tuple(getSortedIds(wf.worklists)), indent),
       wrap(tuple(wf.permissions), indent)
       ))

# States
def dumpStates(wf, aStrIO):

    from zLOG import LOG, INFO
    LOG('#############', INFO, repr([ state.getId() for state in getSortedObjects(wf.states) ]))
    LOG('#############', INFO, repr(wf.initial_state))
    if wf.initial_state not in [ state.getId() for state in getSortedObjects(wf.states) ]:
        raise AttributeError

    aStrIO.write("""\
    # Initial State
    wf.states.setInitialState(%s)

    # State Initialization""" % quote(wf.initial_state))

    for state in getSortedObjects(wf.states):
        dumpState(state, aStrIO)

def dumpState(aState, aStrIO):
    aStrIO.write("""
    sdef = wf.states[%s]
    sdef.setProperties(title=%s,
                       description=%s,
                       transitions=%s)
""" % (quote(aState.getId()),
       wrap(aState.title, indent=30),
       quote(aState.description),
       aState.getTransitions()
       ))
    # Permissions
    for _perm in aState.getManagedPermissions():
        _perm_info = aState.getPermissionInfo(_perm)
        aStrIO.write("    sdef.setPermission(%s, %i,\n"
                     "                       %s)\n"
                     % (quote(_perm), _perm_info['acquired'],
                        wrap(_perm_info['roles'], indent=23)))
    # Assignments to variables
    vars = aState.var_values
    if vars:
        for _id in vars.keys():
            aStrIO.write("    sdef.addVariable(id=%s, value=%s)\n" %
                         (quote(_id), quote(vars.get(_id))))

# Transitions
def dumpTransitions(wf, aStrIO):
    aStrIO.write("""
    # Transition Initialization""")

    for transition in getSortedObjects(wf.transitions):
        dumpTransition(transition, aStrIO)

def dumpTransition(aTransition, aStrIO):
    aStrIO.write("""
    tdef = wf.transitions[%s]
    tdef.setProperties(title=%s,
                       description=%s,
                       new_state_id=%s,
                       trigger_type=%i,
                       script_name=%s,
                       after_script_name=%s,
                       actbox_name=%s,
                       actbox_url=%s,
                       actbox_category=%s,
                       props=%s,
                       )
""" % (quote(aTransition.getId()),
       wrap(aTransition.title, indent=30),
       quote(aTransition.description),
       quote(aTransition.new_state_id),
       aTransition.trigger_type,
       quote(aTransition.script_name),
       quote(aTransition.after_script_name),
       quote(aTransition.actbox_name),
       quote(aTransition.actbox_url),
       quote(aTransition.actbox_category),
       wrap(getGuardProps(aTransition.getGuard()), indent=30)
        ))
    # Var. assignments
    vars = aTransition.var_exprs
    if vars:
        for _id in vars.keys():
            aStrIO.write("    tdef.addVariable(id=%s, text=%s)\n" %
                         (quote(_id), repr(aTransition.getVarExprText(_id))))

# Variables
def dumpVariables(wf, aStrIO):
    aStrIO.write("""
    # State Variable
    wf.variables.setStateVar(%s)

    # Variable Initialization""" % quote(wf.state_var))

    for variable in getSortedObjects(wf.variables):
        dumpVariable(variable, aStrIO)

def dumpVariable(aVariable, aStrIO):
    aStrIO.write("""
    vdef = wf.variables[%s]
    vdef.setProperties(description=%s,
                       default_value=%s,
                       default_expr=%s,
                       for_catalog=%i,
                       for_status=%i,
                       update_always=%i,
                       props=%s)
""" % (quote(aVariable.getId()),
       quote(aVariable.description),
       quote(aVariable.default_value),
       quote(aVariable.getDefaultExprText()),
       aVariable.for_catalog,
       aVariable.for_status,
       aVariable.update_always,
       wrap(getGuardProps(aVariable.getInfoGuard()), indent=30)
       ))

# Worklists
def dumpWorklists(wf, aStrIO):
    aStrIO.write("""
    # Worklist Initialization""")

    for worklist in getSortedObjects(wf.worklists):
        dumpWorklist(worklist, aStrIO)

def dumpWorklist(aWorklist, aStrIO):
    """Dumps a worklist.
    """
    aStrIO.write("""
    ldef = wf.worklists[%s]
    ldef.setProperties(description=%s,
                       actbox_name=%s,
                       actbox_url=%s,
                       actbox_category=%s,
                       props=%s)
""" % (quote(aWorklist.getId()),
       quote(aWorklist.description),
       quote(aWorklist.actbox_name),
       quote(aWorklist.actbox_url),
       quote(aWorklist.actbox_category),
       wrap(worklistProps(aWorklist, getGuardProps(aWorklist.getGuard())),
            indent=30)
        ))

def worklistProps(aWorklist, props):
    """Updates the props dictionary with key/value pairs stored in the
    worklist's var_match.
    """
    if props is None: props={}
    for key in aWorklist.var_matches.keys():
        props['var_match_'+key]=';'.join(aWorklist.var_matches.get(key))
    return props

# Guard
def getGuardProps(aGuard):
    """Inspects aGuard's properties and returns them in a dictionary.
    """
    props={}
    if (aGuard.getPermissionsText()):
        props['guard_permissions'] = aGuard.getPermissionsText()
    if (aGuard.getRolesText()):
        props['guard_roles'] = aGuard.getRolesText()
    if (aGuard.getExprText()):
        props['guard_expr'] = aGuard.getExprText()
    if props:
        return props
    else:
        return None

def dumpScripts(wf, aStrIO):
    """Dump all scripts
    """
    scripts = getSortedObjects(wf.scripts)
    if scripts:
        aStrIO.write('\n    # Script Initialization')
    for script in scripts:
        meta_type = script.meta_type
        if meta_type == 'External Method':
            dumpExternalMethod(script, aStrIO)
            continue
        elif meta_type != 'Script (Python)':
            continue

        proxy_roles = []
        for role in script.valid_roles():
            if script.manage_haveProxy(role):
                proxy_roles.append(role)
        script_data = {
            'id': quote(script.id),
            'title': quote(script.title),
            'body': script.read(),
            'params': quote(script.params()),
            'proxy_roles': repr(tuple(proxy_roles)),
            'owner': quote(script.getOwner().getId()),
            }
        aStrIO.write("""\n
    wf.scripts._setObject(%(id)s, PythonScript(%(id)s))
    script = wf.scripts[%(id)s]
    script.write('''%(body)s''')
    setattr(script, 'title', %(title)s)
    # proxy roles are useless in workflows and should be avoided
    # the following code is commented by default, and it may not work properly
    #setattr(script, '_proxy_roles', %(proxy_roles)s)
    setattr(script, '_owner', %(owner)s)\n\n""" % script_data)

def dumpExternalMethod(script, aStrIO):
    """
    dump code for External Method
    """
    data = {'id': script.id,
            'title': script.title,
            'module': script.module(),
            'function': script.function()
            }

    aStrIO.write("""
    wf.scripts._setObject('%(id)s',
                          ExternalMethod('%(id)s', '%(title)s',
                                         '%(module)s', '%(function)s'))
    """ % data)
    
# Enrobage
def writeHeader(wf, aStrIO):
    """Writes the beginning of of python script, and indicates which
    scripts should be backed up, if applicable.
    """
    aStrIO.write("""\
# Generated by dumpDCWorkflow.py written by Sebastien Bigaret
# Original workflow id/title: %s/%s
# Date: %s
#
# WARNING: this dumps does NOT contain any scripts you might have added to
# the workflow, IT IS YOUR RESPONSABILITY TO MAKE BACKUPS FOR THESE SCRIPTS.
#
""" % (wf.id, wf.title, DateTime()) )

    _scripts = wf.scripts.objectValues()
    if len(_scripts):
        aStrIO.write("# The following scripts have been detected and should be backed up:\n")
        for _script in _scripts:
            _id = _script.id
            if callable(_id): _id=_id()
            aStrIO.write("# - %s (%s)\n" % (_id, _script.meta_type))
    else:
        aStrIO.write("# No script detected in this workflow\n")

    aStrIO.write("#\n")
    aStrIO.write(triple_quotes+"""
Programmatically create a workflow type.
"""+triple_quotes+'\n')
    aStrIO.write("""__version__ = "$Id: dumpDCWorkflow.py 25723 2006-07-04 08:41:22Z b_mathieu $"\n""")

def writeFooter(wf, aStrIO):
    """Write the footer.
    """
    aStrIO.write("""
def create_%s(id):
    \"\"\"Create, setup and return the workflow.
    \"\"\"
    ob = DCWorkflowDefinition(id)
    setup_%s(ob)
    return ob

addWorkflowFactory(create_%s,
                   id=%s,
                   title=%s)
""" % (wf.id, wf.id, wf.id, quote(wf.id), wrap(wf.title, indent=26)))
